// vporshe.cc --- The Porshe class with visiting behavior.

// Author:        Didier Verna <didier@lrde.epita.fr>
// Maintainer:    Didier Verna <didier@lrde.epita.fr>
// Created:       Wed Oct 15 10:05:47 2008
// Last Revision: Mon Apr  6 16:30:37 2009


// Commentary:

// Contents management by FCM version 0.1.

// In this version, the Porshe hierarchy is equipped with generic visiting
// behavior, and the painting and recycling actions involve no more
// modification to the original classes.


// Code:

#include <iostream>
#include <ostream>
#include <vector>

struct Wheel;
struct Body;
struct Engine;
struct Porshe;

// A PorsheVisitor must implement the methods to visit all
// VisitablePorsheComponents.
struct PorsheVisitor
{
  PorsheVisitor () {};
  virtual ~PorsheVisitor () {};

  virtual void visit (Wheel&) = 0;
  virtual void visit (Body&) = 0;
  virtual void visit (Engine&) = 0;
  virtual void visit (Porshe&) = 0;
};

// A VisitablePorsheComponent implements a method for letting a visitor
// perform its task.
struct VisitablePorsheComponent
{
  VisitablePorsheComponent () {};
  virtual ~VisitablePorsheComponent () {};

  virtual void accept (PorsheVisitor&) = 0;
};

// A wheel
struct Wheel : public VisitablePorsheComponent
{
  Wheel () { std::cout << "Created a wheel." << std::endl; };

  virtual void accept (PorsheVisitor& visitor)
  {
    visitor.visit (*this);
  };
};

// A body with 4 wheels
struct Body : public VisitablePorsheComponent
{
  Body ()
  {
    for (int i=0; i<4; i++)
      _wheels.push_back (Wheel());
    std::cout << "Created a body." << std::endl;
  };

  virtual void accept (PorsheVisitor& visitor)
  {
    visitor.visit (*this);
    for (std::vector<Wheel>::iterator i = _wheels.begin ();
	 i != _wheels.end ();
	 i++)
      i->accept (visitor);
  };

  std::vector<Wheel> _wheels;
};

// An engine
struct Engine : public VisitablePorsheComponent
{
  Engine () { std::cout << "Created an engine." << std::endl; };

  virtual void accept (PorsheVisitor& visitor)
  {
    visitor.visit (*this);
  };
};

// A Porshe with a 4 wheels body and an engine
struct Porshe : public VisitablePorsheComponent
{
  Porshe () { std::cout << "Created a porshe." << std::endl; };

  virtual void accept (PorsheVisitor& visitor)
  {
    visitor.visit (*this);
    _engine.accept (visitor);
    _body.accept (visitor);
  };

  Body _body;
  Engine _engine;
};


// A painting visitor
struct PaintPorsheVisitor : public PorsheVisitor
{
  PaintPorsheVisitor () {};
  virtual ~PaintPorsheVisitor () {};

  virtual void visit (Wheel& wheel)
  {
    std::cout << "Painting a wheel." << std::endl;
  };
  virtual void visit (Body& body)
  {
    std::cout << "Painting the body." << std::endl;
  };
  virtual void visit (Engine& engine)
  {
    std::cout << "Painting the engine." << std::endl;
  };
  virtual void visit (Porshe& porshe)
  {
    std::cout << "Painting the porshe." << std::endl;
  };
};

// A recycling visitor
struct RecyclePorsheVisitor : public PorsheVisitor
{
  RecyclePorsheVisitor () {};
  virtual ~RecyclePorsheVisitor () {};

  virtual void visit (Wheel& wheel)
  {
    std::cout << "Recycling a wheel." << std::endl;
  };
  virtual void visit (Body& body)
  {
    std::cout << "Recycling the body." << std::endl;
  };
  virtual void visit (Engine& engine)
  {
    std::cout << "Recycling the engine." << std::endl;
  };
  virtual void visit (Porshe& porshe)
  {
    std::cout << "Recycling the porshe." << std::endl;
  };
};


int main (int argc, char *argv[])
{
  Porshe porshe;

  PaintPorsheVisitor painter;
  porshe.accept (painter);

  RecyclePorsheVisitor recycler;
  porshe.accept (recycler);
}

// vporshe.cc ends here
